@@ -1,7 +1,7 @@
 /*
  *	The PCI Utilities -- List All PCI Devices
  *
- *	Copyright (c) 1997--2006 Martin Mares <mj@ucw.cz>
+ *	Copyright (c) 1997--2007 Martin Mares <mj@ucw.cz>
  *
  *	Can be freely distributed and used under the terms of the GNU GPL.
  */
@@ -59,7 +59,9 @@ static struct pci_access *pacc;
  *  use alloca() much.
  */
 
-#ifndef __GNUC__
+#ifdef __GNUC__
+#include <alloca.h>
+#else
 #undef alloca
 #define alloca xmalloc
 #endif
@@ -75,6 +77,7 @@ struct device {
 };
 
 static struct device *first_dev;
+static int seen_errors;
 
 static int
 config_fetch(struct device *d, unsigned int pos, unsigned int len)
@@ -96,7 +99,7 @@ config_fetch(struct device *d, unsigned int pos, unsigned int len)
 	d->config_bufsize *= 2;
       d->config = xrealloc(d->config, d->config_bufsize);
       d->present = xrealloc(d->present, d->config_bufsize);
-      bzero(d->present + orig_size, d->config_bufsize - orig_size);
+      memset(d->present + orig_size, 0, d->config_bufsize - orig_size);
     }
   result = pci_read_block(d->dev, pos, d->config + pos, len);
   if (result)
@@ -114,7 +117,7 @@ scan_device(struct pci_dev *p)
   if (!pci_filter_match(&filter, p))
     return NULL;
   d = xmalloc(sizeof(struct device));
-  bzero(d, sizeof(*d));
+  memset(d, 0, sizeof(*d));
   d->dev = p;
   d->config_cached = d->config_bufsize = 64;
   d->config = xmalloc(64);
@@ -122,8 +125,9 @@ scan_device(struct pci_dev *p)
   memset(d->present, 1, 64);
   if (!pci_read_block(p, 0, d->config, 64))
     {
-      fprintf(stderr, "Unable to read the standard header from device "
-	      "%04x:%02x:%02x.%d\n", p->domain, p->bus, p->dev, p->func);
+      fprintf(stderr, "lspci: Unable to read the standard configuration space header of device %04x:%02x:%02x.%d\n",
+	      p->domain, p->bus, p->dev, p->func);
+      seen_errors++;
       return NULL;
     }
   if ((d->config[PCI_HEADER_TYPE] & 0x7f) == PCI_HEADER_TYPE_CARDBUS)
@@ -1241,6 +1245,9 @@ show_express(struct device *d, int where, int cap)
     case PCI_EXP_TYPE_PCI_BRIDGE:
       printf("PCI/PCI-X Bridge");
       break;
+    case PCI_EXP_TYPE_PCIE_BRIDGE:
+      printf("PCI/PCI-X to PCI-Express Bridge");
+      break;
     default:
       printf("Unknown type");
   }
@@ -1998,17 +2005,17 @@ grow_tree(void)
 	  b = xmalloc(sizeof(struct bridge));
 	  b->domain = d->dev->domain;
 	  if (ht == PCI_HEADER_TYPE_BRIDGE)
-	    {
-	      b->primary = get_conf_byte(d, PCI_CB_PRIMARY_BUS);
-	      b->secondary = get_conf_byte(d, PCI_CB_CARD_BUS);
-	      b->subordinate = get_conf_byte(d, PCI_CB_SUBORDINATE_BUS);
-	    }
-	  else
 	    {
 	      b->primary = get_conf_byte(d, PCI_PRIMARY_BUS);
 	      b->secondary = get_conf_byte(d, PCI_SECONDARY_BUS);
 	      b->subordinate = get_conf_byte(d, PCI_SUBORDINATE_BUS);
 	    }
+	  else
+	    {
+	      b->primary = get_conf_byte(d, PCI_CB_PRIMARY_BUS);
+	      b->secondary = get_conf_byte(d, PCI_CB_CARD_BUS);
+	      b->subordinate = get_conf_byte(d, PCI_CB_SUBORDINATE_BUS);
+	    }
 	  *last_br = b;
 	  last_br = &b->chain;
 	  b->next = b->child = NULL;
@@ -2312,7 +2319,7 @@ map_the_bus(void)
       pacc->method == PCI_ACCESS_DUMP)
     printf("WARNING: Bus mapping can be reliable only with direct hardware access enabled.\n\n");
   bus_info = xmalloc(sizeof(struct bus_info) * 256);
-  bzero(bus_info, sizeof(struct bus_info) * 256);
+  memset(bus_info, 0, sizeof(struct bus_info) * 256);
   if (filter.bus >= 0)
     do_map_bus(filter.bus);
   else
@@ -2405,5 +2412,5 @@ main(int argc, char **argv)
     }
   pci_cleanup(pacc);
 
-  return 0;
+  return (seen_errors ? 2 : 0);
 }
